import numpy as np
import tensorflow as tf

from scipy.optimize import minimize
from scipy.optimize import Bounds

from skimage import io, color
from PIL import Image

try:
    from utils import *
except ImportError:
    from .utils import *
    
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

def convert_constraints_and_weights_data( pixel_lum_constraints, curve_lum_constraints, W, N, num_p ):
    pl_cons, cl_cons = [], []
    
    # convert luminance constraints from pixels
    for c in pixel_lum_constraints:
        pl_cons.append( (int(c[0]*N), c[1]) )
    
    # convert luminance constraints from curves themselves
    for c in curve_lum_constraints:
        select_palette = c[0]
        cl_cons.append( (int(c[1]*N) + N*select_palette, c[2]) )
        
    if W.shape[0] != 0:
        W[-1, :] /= 100
        W = W / np.sum(W, axis=0)
    else:   # if currently no pixel weigh constraints, just make it #p-by-1 zero matrix
        W = np.zeros( (num_p, 1) ) 
    return pl_cons, cl_cons, W

def grad( n ):
    G = np.zeros( ( n-1, n ) )
    for i in range( 0, n-1 ):
        G[i,i+1] = 1
        G[i,i] = -1
    return G

def linear_sys_solve( constraints, curve_constraints, W, N, w_c ):
    '''
    Given:
        `constraints`: A list of luminance constraints, i.e. [ (int, 0.2), (int, 0.3), ... ]
        `curve_constraints`: A list of curve constraints, i.e. [ (int, 0.2), (int, 0.3), ... ]
        `W`: A #p-by-#c array of weights at constraints
        `N`: Number of samples (100 is enough)
        `w_c`: A scalar of the penalization term
    
    Return:
        `L_`: A N-by-#p array of bilaplacian luminances
    '''
    # identities in kronecker product
    id_p = np.eye( W.shape[0] ) 
    id_n = np.eye( N ) 
    
    # building bilaplacian
    G = grad( N )
    Lap = G.T @ G
    
    # mirroring the laplacian to avoid prefering 0 gradients at endpoints
    Lap[0,:] = 0
    Lap[-1,:] = 0
    B = Lap
    
    # building selection and constraint matrices
    if len( constraints ) == 0: # if no luminance constraints, then set C and S to be zeros
        C = np.zeros( ( N, 1 ) )
        S = np.zeros( ( N, 1 ) )
    else:
        C = np.zeros( ( N, len( constraints ) ) )
        S = np.zeros( ( N, len( constraints ) ) )
    for i, ( Lindex, val ) in enumerate( constraints ):
        C[Lindex, i] = val
        S[Lindex, i] = 1
        
        
    # compute left-hand and right-hand side
    # note: `flatten('F')` is vectorization on columns
    W_idn_kron = np.kron( W, id_n )
    A = np.kron( id_p, 2*B.T@B ) + w_c * W_idn_kron @ np.diag( (2*np.multiply(S,S)).flatten('F') ) @ np.kron( W.T, id_n )
    b = w_c * ( W_idn_kron @ ( np.multiply( 2*C, S ).flatten('F') ) )
    
    # change the rows to identity for known values
    id_p_n = np.eye( W.shape[0] * N )
    zero_lum_ind = [N*i for i in range(W.shape[0])]
    one_lum_ind = [N*(i+1)-1 for i in range(W.shape[0])]
    
    A[zero_lum_ind, :] = id_p_n[zero_lum_ind, :]
    A[one_lum_ind, :] = id_p_n[one_lum_ind, :]
    b[zero_lum_ind] = 0
    b[one_lum_ind] = 1
    
    # placing curve constraints from direct manipulation of curves
    if len( curve_constraints ) != 0:
        curve_cons_ind = [i[0] for i in curve_constraints]
        curve_cons_val = np.array( [i[1] for i in curve_constraints] )
        
        A[curve_cons_ind, :] = id_p_n[curve_cons_ind, :]
        b[curve_cons_ind] = curve_cons_val
        
    x_l = np.linalg.solve( A, b )
    L_ = x_l.reshape( N, W.shape[0], order='F' )
    
    return L_