import numpy as np
import matplotlib.pyplot as plt

from scipy.interpolate import pchip_interpolate
from scipy.optimize import minimize
from scipy.optimize import Bounds


def surrogate_optimization( L0_cons, L_target ):
    '''
    Given:
        `L0_cons`: A shape (#C, ) of array consisting of input luminance constraints
        `L_target`: A shape (#C, ) of array consisting of output luminance constraints

    Return:
        ``
    '''
    # end points
    fixed_pts = np.array( [0, 1] )
    
    # samples for computing approximate integrals
    n = np.linspace(0, 1, 50)
    
    # surrogate constraints s_i
    sur_cons = np.array( [.25, .5, .75] )
    
    # all sorted constraints
    cons_x = np.concatenate( ( sur_cons, fixed_pts, L0_cons ) )
    sort_ind = np.argsort( cons_x )
    cons_x = cons_x[ sort_ind ]
    
    # compute d_si
    d_s = np.ones_like( sur_cons )
    for i in range( d_s.shape[0] ):
        closest_ind = np.argmin( np.abs( sur_cons[i]-L0_cons ) )
        c_s = L0_cons[ closest_ind ]
        c_s_ = L_target[ closest_ind ]
        d_s[i] = ( c_s_ - c_s ) / np.abs( c_s_ - c_s )
    
    # compute d_ni
    d_n = np.ones_like( n )
    for i in range( d_n.shape[0] ):
        closest_ind = np.argmin( np.abs( n[i]-L0_cons ) )
        dir = L_target[ closest_ind ] - L0_cons[ closest_ind ]
        d_n[i] = dir / np.abs( dir )
    
    # set boundary for o_si
    bnds = []
    for i in range( d_s.shape[0] ):
        if d_s[i] == -1:
            bnds.append( ( -sur_cons[i], 0 ) )
        else:
            bnds.append( ( 0, 1-sur_cons[i] ) )
    
    
    ### Optimization losses
    def integral_loss( new_sur_cons ):
        cons_y = np.concatenate( ( new_sur_cons, fixed_pts, L_target ) )
        cons_y = cons_y[ sort_ind ]
        
        # compute integral values for samples
        H_X_n = pchip_interpolate( cons_x, cons_y, n )
        inference = H_X_n - n
        d_n_ = np.sign( H_X_n - n )
        print( d_n_ )
        penal = np.zeros_like( d_n_ )
        for i in range( d_n_.shape[0] ):
            if d_n_[i] != 0 and d_n_[i] * d_n[i] == -1:
                penal[i] = 20
        
        integrals = np.sum( inference*penal )
        return integrals
    
    def total_loss( offsets ):
        #sparsity = np.linalg.norm( offsets )
        sparsity = np.sum( np.abs( offsets ) )
        
        # creating surrogate constraints
        new_sur_cons = sur_cons + offsets
        integrals = integral_loss( new_sur_cons )
        
        return sparsity + 0.05 * integrals
    
    x0 = np.zeros_like( sur_cons )
    res = minimize( total_loss, x0, method='SLSQP', options={'ftol': 1e-6, 'disp': False, 'maxiter': 300}, bounds=bnds )
    print( res )
    
    #print( res.x )
    #print( loss(res.x) )
    sur_cons = sur_cons + res.x
    return sur_cons
    
    
    
def plot_pchip( L0_cons, L_target ):
    L = np.linspace( 0, 1, 500 )
    L_ = pchip_interpolate( L0_cons, L_target, L )
    
    # plot
    plt.plot( L, L_, label = "pchip interpolation" )
    plt.plot( L, L, label = "y=x" )
    
    plt.axvline( x = L0_cons[-2], linewidth = 1.5, color = 'r', label = 'L0', linestyle=':' )
    plt.axhline( y = L_target[-2], linewidth = 1.5, color = 'g', label = 'L0\'', linestyle=':' )
    
    plt.grid()
    plt.legend()
    plt.show()
    
def main():
    L0_cons = np.array([0, .25, .5, .75, 0.8, 1])
    L_target = np.array([0, .25, .5, .75, 0.9, 1])
    
    L_constraints = np.array( [[0.8, 0.9]] )
    sur_offsets = surrogate_optimization( L_constraints[:, 0], L_constraints[:, 1] )
    L_target[1:4] = sur_offsets
    plot_pchip( L0_cons, L_target )
    

if __name__ == '__main__':
    main()












'''
def laplacian_solve():
    from scipy.sparse import diags
    cons = 0.82144326
    
    # create 1D laplace operator
    n = 100
    k = [np.ones(n-1),-2*np.ones(n),np.ones(n-1)]
    offset = [-1,0,1]
    L = diags( k, offset ).toarray()
    
    # create constraint matrices
    s = np.linspace( 0, 1, n )
    C = np.zeros( (3,n) )
    C[0,0] = 1
    C[1,-1] = 1
    C[2, np.argmin(np.abs(s - cons))] = 1
    
    d = np.zeros( (3,1) )
    d[1] = 1
    d[-1] = 0.60319034
    
    # solve linear system
    w = 100
    lhs = L + w*(C.T @ C) 
    rhs = w*(C.T @ d)
    
    ind1 = np.argmin( np.abs(s-0.25) )
    ind2 = np.argmin( np.abs(s-0.5) )
    ind3 = np.argmin( np.abs(s-0.75) )
    def loss( x ):
        lap = np.linalg.norm( lhs @ x - rhs.reshape(-1) )
        sparsity = np.abs(0.25-x[ind1]) + np.abs(0.5-x[ind2]) + np.abs(0.75-x[ind3])
        return 200*lap + sparsity
    
    x0 = np.zeros_like( s )
    bounds = Bounds( 0, 1 )
    res = minimize( loss, x0, method='SLSQP', options={'ftol': 1e-6, 'disp': False, 'maxiter': 300}, bounds=bounds )
    
    plt.plot( s, np.linalg.solve(lhs, rhs), label = "Laplacian" )
    plt.plot( s, s, label = "y=x" )
    plt.axvline( x = cons, linewidth = 1.5, color = 'r', label = 'L0', linestyle=':' )
    plt.axhline( y = d[-1], linewidth = 1.5, color = 'g', label = 'L0\'', linestyle=':' )
    plt.grid()
    plt.legend()
    plt.show()
'''